// system.h



enum alphabet_type         {GROUP, MONOID};
enum rewriting_strategy    {FROM_LEFT, GLOBAL};
enum knuth_bendix_strategy {BASIC, TWO_STEP};

enum order_type            {LNLX, WTLNLX, WRLNLX, WRWTLNLX,
                            RWRLNLX,RWRWTLNLX};
                            
enum naming_convention     {PREFIX, LETTERS, NO_NAME};
enum numbering_convention  {a_b, a_A, NO_NUMB};
enum inverse_convention    {CASE, EXPONENT, NO_INV};
enum convention            {CONSTANT, NONCONSTANT,RNONCONSTANT};



/////////////////////////////////////////////////////////////////
//
//                       rewriting_system
//
/////////////////////////////////////////////////////////////////

class rewriting_system {
  vector_char    system_file;
  alphabet_type  a_type;
  bool           patterns;  // TRUE if patterns in use.
  order_type     o_type;
  bool           use_heuristics;
  long           n_gens;    // Number of generators.
  long           t_gens;    // Total number of monoid generators.
                            // The same as n_gens for monoid
                            // alphabets and 2*n_gens for group
                            // alphabets.
  long           n_xgens;   // Includes the generators used in
                            // patterns.
  long           t_xgens;   // Includes the generators used in
                            // patterns.
  long           n_bits;    // Number of bits needed to pack a
                            // generator or the inverse of a
                            // generator.
  generator     *gens;      // Array of generator information.
  gen_autn       autn;      // Facilitates parsing.
  long           n_sub;     // The number of subsystems
  long           savell;    // Max length of left sides of rules
                            // currently being saved.
  long           savelr;    // Max length of right sides of rules
                            // currently being saved.
  long           bsize;     // Current length of rewriting
                            // buffers.

  rewriting_strategy      rw_strategy;
  knuth_bendix_strategy   kb_strategy;

  bool           rewrite_verbose;   // If true, report on
                                    // rewriting.
  bool           use_patterns;      // If true, use rule patterns
                                    // in rewriting.
  bool           use_broken;        // If true, retain "broken
                                    // rules".
  long           logll;
  external_form  logform;
  ofstream       log_stream;

  long           n_rewrite;
  long           total_input;
  long           n_applications;
  long           total_left;
  long           noverlaps;
  long           nbalance;
  long           nsave;
  long           nsave_total;
  long           nsave_max;
  long           n_rabin_karp;
  long           n_rk_tries;
  long           n_rk_hits;

  // The subsystems are described by various doubly linked lists

  subsystem     *first_subsystem;
  subsystem     *last_subsystem;
  subsystem     *first_save;
  subsystem     *last_save;
  subsystem     *broken_subword_rules;
  subsystem     *broken_prefix_rules;
  
 public:
  rewriting_system(void);

 ~rewriting_system();

  subsystem  *new_subsystem(char *str = NULL);

  void get_system_file(char *fil) { system_file.get_data(fil); }
  alphabet_type get_a_type(void) { return a_type; }

  long          get_n_gens(void)  { return n_gens; }
  long          get_t_gens(void)  { return t_gens; }
  long          get_n_xgens(void) { return n_xgens; }
  long          get_t_xgens(void) { return t_xgens; }
  long          get_n_bits(void)  { return n_bits; }
  long          get_n_sub(void)   { return n_sub;  }

  bool get_use_heuristics() { return use_heuristics; }
  void set_use_heuristics(bool value)
    { use_heuristics = value; }

  void input(char * file, bool check);
  void input(char * directory, istream &stream, bool prompt, bool check);

  long inverse(long g) { return gens[g].get_inv_numb(); }
  long gen_number(char *str);
  void gen_name(char *nam, long i) { gens[i].get_name(nam); }
  long parse(vector_char &vec, vector_long &wrd);
  long parse_6bit(vector_char &vec, long nbits, long len,
                  vector_long &word);
  void add_subsystem( subsystem *subsyst );
  void make_generator_index();
  long read_word( istream &stream, external_form exf,
                  long nbits, vector_long &wrd);
  void summary();

  void set_save( long ll, long lr)
    { savell = ll; savelr = lr; bsize = max(bsize,max(ll,lr) ); }
  void set_bsize( long bsz ) { bsize = bsz; }
  long get_bsize() { return bsize; }
  long get_weight( long g ) { return gens[g].get_weight(); }
  long get_level( long g ) { return gens[g].get_level(); }
  long word_weight(vector_long &wrd);
  bool orient_rule(rule &rul);
  void balance_rule(rule &rul);

  void save_rule(rule &rul);
  void save_broken_rule(rule &rul);
  
  void set_rewrite_verbose( bool val )
    { rewrite_verbose = val; }

  bool get_use_patterns() { return use_patterns; }

  void set_use_patterns( bool val )
    { use_patterns = val; }

  bool get_use_broken() { return use_broken; }
  void set_use_broken( bool val );

  void ask_rewrite_verbose(void);

  long (rewriting_system::*base_order)
            (vector_long &u, vector_long &v);

  order_type  compare;

//  long (rewriting_system::*compare)
//            (vector_long  &u, vector_long  &v);

  long comp(vector_long &u, vector_long &v);
//    { return (this->*compare)(u,v); }

  long lnlx(vector_long &u, vector_long &v);

  long wtlnlx(vector_long &u, vector_long &v);

  long wreath(
     vector_long &u,
     vector_long &v
  );

  long wrlnlx(vector_long &u, vector_long &v)
//    { base_order = NULL; return wreath(u, v); }
    { base_order = &lnlx; return wreath(u, v); }

  long wrwtlnlx(vector_long &u, vector_long &v)
//    { base_order = NULL; return wreath(u, v); }
    { base_order = &wtlnlx; return wreath(u, v); }

  long rwreath(
     vector_long &u,
     vector_long &v
  );

  long rwrlnlx(vector_long &u, vector_long &v)
//    { base_order = NULL; return rwreath(u, v); }
    { base_order = &lnlx; return rwreath(u, v); }

  long rwrwtlnlx(vector_long &u, vector_long &v)
//    { base_order = NULL; return rwreath(u, v); }
    { base_order = &wtlnlx; return rwreath(u, v); }

  void level_decomposition(vector_long &u, vector_long &v,
    vector_long &ubuff, vector_long &vbuff);

  void level_sequence(
    vector_long  &u,
    long          lc,
    long          lu,
    vector_long  &ubuff
  );

  void rlevel_decomposition(vector_long &u, vector_long &v,
    vector_long &ubuff, vector_long &vbuff);

  void rlevel_sequence(
    vector_long  &u,
    long          lc,
    long          lu,
    vector_long  &ubuff
  );

  void set_o_type(order_type otp);

  void ( rewriting_system::*pRewrite)(
           bool            subword_only,
           vector_long    &input,
           vector_long    &output,
           rule_reference  exclude,
           bool           &changed,
           bool           &reduced
  );
  
  void rewrite(rule &rul, rule_reference exclude,
       bool &changed, bool &reduced);

  void rewrite_from_left(
           bool            subword_only,
           vector_long    &input,
           vector_long    &output,
           rule_reference  exclude,
           bool           &changed,
           bool           &reduced
  );

  void rewrite_global(
           bool            subword_only,
           vector_long    &input,
           vector_long    &output,
           rule_reference  exclude,
           bool           &changed,
           bool           &reduced
  );

  void set_rw_strategy(rewriting_strategy rws)
    { rw_strategy = rws;
//      if (rws == FROM_LEFT) pRewrite = NULL;
//      else pRewrite = NULL;
    }
  
/*
  bool (rewriting_system::*pKnuth_Bendix)(
    long      maxlev,
    long      maxlen,
    long     &num_iter,
    long      prevlen,
    long      flag
  );
*/

  bool kb_basic(
    long      maxlev,
    long      maxlen,
    long     &num_iter,
    long      prevlen,
    long      flag
  );
  
  bool kb_two_step(
    long      maxlev,
    long      maxlen,
    long     &num_iter,
    long      prevlen,
    long      flag
  );
  
  bool kb_random_two_step(
    long      maxlen,
    rule_type rtype,
    long      ntrials
  );
  
  void set_kb_strategy(knuth_bendix_strategy kbs)
    { return; }

/*
    { kb_strategy = kbs;
      if (kbs == BASIC) pKnuth_Bendix = kb_basic;
      else pKnuth_Bendix = kb_two_step; }
*/

  void initialize_save_list(
    rule_type  rtp,
    long       enm,
    long       ell
  );

  bool reduce_all_subsystems(
    rule_type  rtp
  );

  void heuristics(
    rule  &rul
  );

  void heuristics(void);

  bool reduce_save_list(void);

  void append_save_list(subsystem *subsyst);

  void overlap(
    subsystem  *subsyst1,
    subsystem  *subsyst2,
    long        start_orig,
    long        minlen,
    long        maxlen,
    subsystem **avoid,
    long        navoid,
    long        maxlev,
    long       &neworig,
    long       &nnew,
    bool       &successful
  );

  void random_overlap(
    subsystem  *subsyst1,
    subsystem  *subsyst2,
    long        maxlen,
    subsystem **avoid,
    long        navoid,
    double     *count
  );

  void remove_subsystem( subsystem *subsyst );

  long compare_packed_rules( sort_type stp, packed_rule v_rule,
                            packed_rule w_rule);

  long compare_lex(word_reference u_ref, word_reference v_ref);
  void write(char *file_name);
  
  void print_word(
    ostream       &stream,
    external_form  ext_form,
    vector_long   &word
  );
  
  void print_inverse( fstream &stream, generator &g )
    { print_word( stream, NAMES, g.inverse); }
  
  void print_counts();

  bool check_rule( rule &rul ) {
    if ( rul.left.get_length() <= savell &&
         rul.right.get_length() <= savelr   )
      return TRUE;
    else return FALSE;
  }

  void save_inverses();

  void details( long  i );

  long probe(
    bool    subword_only,
    long    ntrials,
    long    depth
  );

  void rewrite_rules(void);

  void add_powers( long exponent, long maxlen, long firstg,
                   long lastg );

  void add_engel( long engelnum, long firstg, long lastg,
                  long maxlen);

  void add_random_powers(
    long ntrials,
    long exponent,
    long maxlen,
    long firstg,
    long lastg
  );

  void add_subword_powers(
    long exponent
  );

  void rewrite_words(void);

  void rewrite_words(
    istream         &instream,
    external_form    ext,
    long             nbits,
    bool             subword_only,
    bool             stop_on_empty,
    ostream         &outstream
  );

  void log_on(long maxlen, char *file, external_form  form);
  void log_off(void);
  void promote(rule &rul);

  void cancel();
  void append_inverse(vector_long &u, vector_long &v);
  void append_inverse(vector_long &u, long  g);
  void append_power(vector_long &u, vector_long &v, long exp);

  long decode(vector_char &buff, long &index, vector_long &word);
  long get_generator(vector_char &buff, long &index, long &g);

  void add_instances(long subno, long maxill, long maxilr,
                     long maxlen, long firstg, long lastg);

  void rule_instances(rule &rul, long maxill, long maxilr,
                      long maxlen, long firstg, long lastg,
                      long &nfail );

  void memory();

  void rk_update( long tries, long hits);

  long word_level( vector_long  &wrd );

  void write_rules(
    long          numsub,
    char*         file_name,
    external_form extfrm,
    ulong         llmin,
    ulong         llmax,
    ulong         lrmin,
    ulong         lrmax,
    ulong         genmin,
    ulong         genmax
  );

  void reidemeister_schreier(void);

  void conjugate( long num_iter);

  void count(vector_long &wrd, long maxlen);

  void countg(vector_long &wrd, long maxlen);

  void countx(vector_long &wrd, long maxlen);

  void count_subgroup(void);

  void add_subs(long maxill, long maxilr, long maxtll,
                long maxtlr, long maxlen, long firstg,
                long lastg);

  void set_nsave_max(long nsm) {
    nsave_max = nsm;
  }

  bool abort() {
    return nsave_max > 0 && nsave_total > nsave_max;
  }

  void get_nonobvious(void);

  void geodesics(vector_long &wrd, long maxlen);

  void shift_rules(
    long  llmax,
    long  lrmax,
    long  firstg,
    long  shift
  );
};



extern void set_hybrid_factor(double factor);
void get_integer(vector_char &buff, long &index, long &n);

